/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 by The Geometry Center.
 *  This file is part of SaVi.  SaVi is free software;
 *  you can redistribute it and/or modify it only under
 *  the terms given in the file COPYRIGHT which you should
 *  have received along with this file.  SaVi may be
 *  obtained by emailing savi@geom.umn.edu.
 *
 *****************************************************
 *  
 * fisheye.c
 *
 */
#include <stdlib.h>
#include <math.h>

#include "sats.h"
#include "constants.h"
#include "tk.h"
#include "stats_utils.h"
#include "tkPlot.h"
#include "globals.h"
#include "tcl_utils.h"
#include "savi.h"

static int fisheye_flag = FALSE;
static int fisheye_vapors_flag = FALSE;
static int fisheye_reverse_view_flag = FALSE;
static int fisheye_print_inview_flag = FALSE;
static double lat, lon;
static TkPlot *plotPtr = NULL;


/*
 * fisheye_on.c
 *
 * Set flags to indicate that fisheye should be displayed.
 * Call fisheye_display to get updated display.
 * (This is repeatedly called during the sats_update cycle.)
 */
char *
fisheye_on_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite_list sl = pconstellation->satellites;
  CentralBody *pcb = pconstellation->pcb;
  static char fisheye_widget[] = ".fisheye.cmd.p";

  /* set variables so that fisheye will be displayed */
  fisheye_flag = TRUE;

  /* get plotPtr */
  plotPtr = (TkPlot *) get_ClientData(fisheye_widget);

  /* display the fisheye view */
  fisheye_display(sl, TRUE, pcb);

  return EMPTY_str;
}

char *
fisheye_off_cmd(int argc, char *argv[])
{
  fisheye_flag = FALSE;
  return EMPTY_str;
}

char *
fisheye_vapors_on_cmd(int argc, char *argv[])
{
  fisheye_vapors_flag = TRUE;
  return EMPTY_str;
}

char *
fisheye_vapors_off_cmd(int argc, char *argv[])
{
  fisheye_vapors_flag = FALSE;
  return EMPTY_str;
}

char *
fisheye_reverse_view_on_cmd(int argc, char *argv[])
{
  fisheye_reverse_view_flag = TRUE;
  return EMPTY_str;
}

char *
fisheye_reverse_view_off_cmd(int argc, char *argv[])
{
  fisheye_reverse_view_flag = FALSE;
  return EMPTY_str;
}

char *
fisheye_print_inview_on_cmd(int argc, char *argv[])
{
  fisheye_print_inview_flag = TRUE;
  return EMPTY_str;
}

char *
fisheye_print_inview_off_cmd(int argc, char *argv[])
{
  fisheye_print_inview_flag = FALSE;
  return EMPTY_str;
}

/*
 * fisheye_set
 *
 * Sets the lat and lon for the fisheye display and redisplays it
 */
char *
fisheye_set_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite_list sl = pconstellation->satellites;
  CentralBody *pcb = pconstellation->pcb;
  double new_lat = (double) atoi(argv[2]);
  double new_lon = (double) atoi(argv[4]); /* was 3 gjl */
  const char format1[] = "set fisheye_lat_lon \"%d %.4f %c  %d %.4f %c\"";
  char cmd[sizeof(format1)+20];
  int lat_deg, lon_deg;
  char latNS, lonEW;
  double lat_min, lon_min;

  new_lat = (new_lat>=0) ? new_lat+atof(argv[3])/60.0 :
    new_lat-atof(argv[3])/60.0;
  new_lon = (new_lon>=0) ? new_lon+atof(argv[5])/60.0 :
    new_lon-atof(argv[5])/60.0;

  lat = new_lat;
  lon = new_lon;

  /* display lat lon */
  if (lat >= 0)
    {
      latNS = 'N';
      lat_deg = (int) floor(lat);
      lat_min = (lat - lat_deg)*60.0;
    }
  else
    {
      latNS = 'S';
      lat_deg = (int) floor(-lat);
      lat_min = (-lat-lat_deg)*60.0;
    }
  if (lon >= 0)
    {
      lonEW = 'E';
      lon_deg = (int) floor(lon);
      lon_min = (lon - lon_deg)*60.0;
    }
  else
    {
      lonEW = 'W';
      lon_deg = (int) floor(-lon);
      lon_min = (-lon-lon_deg)*60.0;
    }
  sprintf(cmd,format1,lat_deg, lat_min, latNS, lon_deg, lon_min, lonEW);
  tcl_script(cmd);

  /* displaye sky view */
  fisheye_display(sl, TRUE, pcb);
  return EMPTY_str;
}

/*
 * fisheye_display
 *
 * update the Tk image
 */
void
fisheye_display(const Satellite_list SL, int force, const CentralBody *pcb)
{
  unsigned int r = FISHEYE_SIZE/2-1;
  int offset = r+2;
  int bgnd_color = -1, ring_color = -1;
  int sat_color, special_sat_color, angle_ring_color;
  float coverage_pct = (90.0-coverage_angle)/90.0;

  if (bgnd_color < 0)
  {
    bgnd_color = Tk_PlotColorLookup("white");
    ring_color = Tk_PlotColorLookup("grey");
    sat_color = Tk_PlotColorLookup("red");
    angle_ring_color = Tk_PlotColorLookup("green");
    special_sat_color = Tk_PlotColorLookup("yellow");
  }

  if ((fisheye_flag == FALSE) && (force == FALSE)) return;
  if (plotPtr == NULL) return;

  if (fisheye_vapors_flag == FALSE)
  {
    /* clear plot window */
    Tk_PlotClear(plotPtr);

    /* draw bullseye */
    Tk_PlotSymbol(plotPtr, offset, offset, bgnd_color, DISK, r);
    Tk_PlotSymbol(plotPtr, offset, offset, ring_color, CIRCLE, r);
    Tk_PlotSymbol(plotPtr, offset, offset, ring_color, CIRCLE, r*2/3);
    Tk_PlotSymbol(plotPtr, offset, offset, ring_color, CIRCLE, r/3);
    Tk_PlotSymbol(plotPtr, offset, offset, ring_color, POINT, 0);
    Tk_PlotSymbol(plotPtr, offset, offset, angle_ring_color, CIRCLE,
		  (unsigned int) (coverage_pct*r));
  }
  
  /* plot satellite positions */
  {
    CartesianCoordinates s;
    double sky[2];
    double t = current_time();
    Satellite_list sl=SL;
    char plotStr[4];
    float look_start, look_dur;
    const float SECSPERMIN = 60.0;

    while (NULL != sl)
    {
	  oe_time_to_geocentric(&s, t, &(sl->s->oe), pcb);
	  if ( sat_to_fisheye(&s, lat, lon, pcb, t, sky) == TRUE )
	  {
	    if (fisheye_reverse_view_flag)
	    {
	      sky[0] = - sky[0];
	    }
	    sprintf(plotStr, "%d", sl->s->gs_look_dur);
	    Tk_PlotString(plotPtr, (int) (offset+r*sky[0]), (int) 
				(offset-r*sky[1]), sat_color, plotStr, 9);
 		if (sl->s->tag == 1)
	      Tk_PlotSymbol(plotPtr, (int) (offset+r*sky[0]), (int) 
				(offset-r*sky[1]), special_sat_color, CROSSHAIR, 5);
		else
	      Tk_PlotSymbol(plotPtr, (int) (offset+r*sky[0]), (int) 
				(offset-r*sky[1]), sat_color, CROSSHAIR, 5);
	    if ((sky[0] * sky[0] + sky[1] * sky[1]) < (coverage_pct * coverage_pct))
	    {
	    	sl->s->gs_look_dur++;
	    }
	  }
	  else if (sl->s->gs_look_dur != 0)
	  {
	    sl->s->gs_look_dur--;
	    look_dur = ((float) sl->s->gs_look_dur * delta_t / SECSPERMIN);
	    look_start = sl->s->t / SECSPERMIN - look_dur - 1;
	    if (fisheye_print_inview_flag)
	    {
	      fprintf(stderr, "Sat: %2d, Access Start: %6.0f mins, Duration: %2d mins.\n", sl->s->id, look_start,
	        (int) look_dur);
	    }
	    sl->s->gs_look_dur = 0;
	  }
	  sl=sl->next;
    }
  }

  /* force display */
  Tk_PlotQueueRefresh(plotPtr);
}
