/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 by The Geometry Center.
 *  This file is part of SaVi.  SaVi is free software;
 *  you can redistribute it and/or modify it only under
 *  the terms given in the file COPYRIGHT which you should
 *  have received along with this file.  SaVi may be
 *  obtained by emailing savi@geom.umn.edu.
 *
 *****************************************************
 *  
 * footprints.c
 *
 * routines for footprints in Geomview
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "globals.h"
#include "constants.h"
#include "gv_utils.h"
#include "sats.h"
#include "orbit_utils.h"
#include "savi.h"

static int footprints_on_flag = FALSE;
static int footprints_geom_exists = FALSE;
static int footprints_geom_needs_updating = FALSE;
static int footprints_angle_used = 0;

static void
footprint_size(double *radius, double *height, const Satellite s, const
	  Constellation *pconstellation);
static void
footprint_write_geom(const Satellite s, const Constellation *pconstellation);

/*
 * footprints_on_cmd()
 *
 * Routine to display footprints
 */
char *
footprints_on_cmd(int argc, char *argv[])
{
  Satellite_list sl = get_constellation()->satellites;

  if (TRUE == footprints_on_flag) return EMPTY_str;
  footprints_on_flag = TRUE;
  footprints_geom_needs_updating = TRUE;

  transforms_needed |= (1 << FOOTPRINTS);
  gv_trans_list_create(sl);
  gv_begin();
  while (sl)
    {
      footprint_display(sl->s, get_constellation());
      sl = sl->next;
    }
  gv_end();

  /* remember what angle was used */
  footprints_angle_used = coverage_angle;

  return EMPTY_str;
}

/*
 * footprints_off_cmd()
 *
 * Routine to turn off display of footprints
 */
char *
footprints_off_cmd(int argc, char *argv[])
{
  Satellite_list sl = get_constellation()->satellites;

  if (FALSE == footprints_on_flag) return EMPTY_str;

  transforms_needed &= ~(1 << FOOTPRINTS);

  gv_begin();
  footprints_gv_delete();
  while (sl)
    {
      footprint_delete(sl->s);
      sl = sl->next;
    }
  footprints_geom_needs_updating = FALSE;
  footprints_on_flag = FALSE;
  gv_end();

  return EMPTY_str;
}


char *
footprints_set_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite s;

  int n = atoi(argv[2]);
  unsigned int flag = atoi(argv[3]);

  s = constellation_search(pconstellation, n);
  if (s) {
    if (flag) s->can_display_footprint = 1;
    else s->can_display_footprint = 0;
    footprints_geom_needs_updating = TRUE;
  }

  return EMPTY_str;
}


/*
 * footprints_rebuild()
 *
 * Routine to resize all the footprints currently built
 */
void
footprints_rebuild()
{
  Constellation *pconstellation = get_constellation();
  Satellite_list sl = pconstellation->satellites;

  if (FALSE == footprints_on_flag) return;

  gv_begin();
  while (sl) {
    footprint_display(sl->s, pconstellation);
    sl = sl->next;
  }
  gv_end();
  
  /* remember what angle was used */
  footprints_angle_used = coverage_angle;

  footprints_geom_needs_updating = FALSE;
}


/*
 * footprints_gv_delete
 *
 * routine to delete geomview footprints object
 */
void
footprints_gv_delete()
{
  if (TRUE == footprints_geom_exists) {
    gv_begin();
    gv_delete_geom("Footprints");
    gv_end();
    footprints_geom_exists = FALSE;
  }
}


/*
 * footprints_relocate
 *
 * Called to update the footprint when the satellites have all been moved
 *
 */
void
footprints_relocate(const Constellation *pconstellation)
{
  Satellite_list sl = pconstellation->satellites;
  int angle_changed = (footprints_angle_used != coverage_angle);

  if (FALSE == footprints_on_flag) return;
  
  while (sl) {
    if (angle_changed || (sl->s->oe.e > 0)) {
       footprint_write_geom(sl->s, pconstellation);
    }
    sl = sl->next;
  }

  /* remember what angle was used */
  footprints_angle_used = coverage_angle;
}

/*
 * footprint_size
 *
 * computes radius and height of footprint for given satellite
 *
 */
static void
footprint_size(double *radius, double *height, const Satellite s, const
	  Constellation *pconstellation)
{
  CentralBody *pcb = pconstellation->pcb;
  double r, theta, c, si, temp;

  r = s->x_S.r / pcb->radius;
  if (r <= 1.0) {
    /* satellite inside central body! */
    *radius = 0;
    *height = 0;
    return;
  }

  /* coverage_angle is global variable (in degrees) */
  theta = coverage_angle*DEG_TO_RAD;
  c = cos(theta);
  si = sin(theta);
  if (MASK_ELEVATION == get_coverage_type()) {
    temp = sqrt(r*r-c*c);
    *radius = c*(temp-si)/r;
    *height = temp*(temp-si)/r;
  } else {
    /* SATELLITE_CONE */
    if (r*si < 1) {
      /* coverage cone intersects earth */
      if (si == 0) {
	/* trivial case - theta = 0 */
	*radius = 0.0;
	*height = r-1.0;
      } else {
	*radius = sin(asin(r*si)-theta);
	*height = (*radius)*c/si;
      }
    } else {
      /* coverage cone does not intersect earth */
      *radius = sqrt(1-1/(r*r));
      *height = r-1/r;
    }
  }
}

/*
 * footprint_write_geom
 *
 * Write footprint geom for a satellite
 */
static void
footprint_write_geom(const Satellite s, const Constellation *pconstellation)
{
  static char line[1000];
  int i;
  double t[16];

  for (i=0; i<16; i++) t[i] = 0.0;
  /*
     compute scaling of footprint - 
       a rigid body rotation plus a translation
       will be applied to the footprint to place it so
       that (0,0,0) will be mapped to the position of the satellite
       and (0,0,-1) will point to the origin
  */
  footprint_size(t, t+10, s, pconstellation);
  t[5] = t[0];
  t[15] = 1.0;
  
  /* start of gv footprint description */
  gv_begin();
  sprintf(line, "(read geometry {define footprint_%d { INST transform {%g %g %g %g %g %g %g %g %g %g %g %g %g %g %g %g} geom :footprint_h} } )\n",
	  s->id,
	  t[0], t[1], t[2], t[3], t[4], t[5], t[6], t[7], t[8], t[9], t[10],
	  t[11], t[12], t[13], t[14], t[15]);
  gv_send(line);
  gv_end();
}

/*
 * footprint_display
 *
 * Takes a satellite and displays its footprint
 */
void
footprint_display(Satellite s, const Constellation *pconstellation)
{
  if (FALSE == footprints_on_flag) return;
  footprints_geom_needs_updating = TRUE;

  footprint_write_geom(s, pconstellation);
}




/*
 * footprint_delete
 *
 * Takes a satellite and deletes its footprint
 * ONLY if footprints_on_flag == TRUE
 */
void
footprint_delete(Satellite s)
{
  static char name[] = "footprint_XXXXXXXXXX";

  if (FALSE == footprints_on_flag) return;
  footprints_geom_needs_updating = TRUE;

  gv_begin();
  sprintf(name+5, "%d", s->id);
  gv_delete_handle(name);
  gv_end();
}



void
write_footprints_geom(const Constellation *pconstellation)
{
  Satellite s;
  Satellite_list sl = pconstellation->satellites;
  char buf[256];

  if ((footprints_on_flag) && (footprints_geom_needs_updating == TRUE))
    {
      /* write out new list of all satellites */
      gv_send("(geometry Footprints {LIST\n");
      while (NULL != sl)
	{
	  s = sl->s;
	  if (sl->s->can_display_footprint) {
	    sprintf(buf, "{ INST transform:trans_%d geom:footprint_%d}\n", s->id, s->id);
	    gv_send(buf);
	  }
	  sl = sl->next;
	}
      gv_send("})\n");
      
      footprints_geom_exists = TRUE;
      footprints_geom_needs_updating = FALSE;
    }
}

