/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 by The Geometry Center.
 *  This file is part of SaVi.  SaVi is free software;
 *  you can redistribute it and/or modify it only under
 *  the terms given in the file COPYRIGHT which you should
 *  have received along with this file.  SaVi may be
 *  obtained by emailing savi@geom.umn.edu.
 *
 *****************************************************
 *  
 * links.c
 *
 * routines for inter-satellite links
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "globals.h"
#include "constants.h"
#include "gv_utils.h"
#include "sats.h"
#include "orbit_utils.h"
#include "savi.h"

static int links_on_flag = FALSE;
static int links_geom_exists = FALSE;
static int links_geom_needs_updating = FALSE;

/*
 * links_on_cmd()
 *
 * Routine to display links
 */
char *
links_on_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite_list sl = pconstellation->satellites;

  if (TRUE == links_on_flag) return EMPTY_str;
  links_on_flag = TRUE;
  links_geom_needs_updating = TRUE;

  gv_begin();
  while (sl)
    {
      link_display(sl->s, pconstellation);
      sl = sl->next;
    }
  gv_end();

  return EMPTY_str;
}

/*
 * links_off_cmd()
 *
 * Routine to turn off display of links
 */
char *
links_off_cmd(int argc, char *argv[])
{
  Satellite_list sl = get_constellation()->satellites;

  if (FALSE == links_on_flag) return EMPTY_str;

  gv_begin();
  links_gv_delete();
  while (sl)
    {
      link_delete(sl->s);
      sl = sl->next;
    }
  links_geom_needs_updating = FALSE;
  links_on_flag = FALSE;
  gv_end();

  return EMPTY_str;
}

char *
links_set_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite s;

  int n = atoi(argv[2]);
  unsigned int flag = atoi(argv[3]);

  s = constellation_search(pconstellation, n);
  if (s) {
    if (flag) s->can_display_links = 1;
    else s->can_display_links = 0;
    links_geom_needs_updating = TRUE;
  }

  return EMPTY_str;
}

/*
 * links_rebuild()
 *
 * Routine to resize all the links currently built
 */
void
links_rebuild()
{
  Constellation *pconstellation = get_constellation();
  Satellite_list sl = pconstellation->satellites;

  if (FALSE == links_on_flag) return;

  gv_begin();
  while (sl) {
    link_display(sl->s, pconstellation);
    sl = sl->next;
  }
  gv_end();
  
  links_geom_needs_updating = FALSE;
}


/*
 * links_gv_delete
 *
 * routine to delete geomview links object
 */
void
links_gv_delete()
{
  if (TRUE == links_geom_exists) {
    gv_delete_geom("Links");
    links_geom_exists = FALSE;
  }
}


/*
 * link_display
 *
 * Takes a satellite and displays its link
 */
void
link_display(Satellite s, const Constellation *pconstellation)
{

}




/*
 * link_delete
 *
 * Takes a satellite and deletes its link
 * ONLY if links_on_flag == TRUE
 */
void
link_delete(Satellite s)
{
  static char name[] = "link_XXXXXXXXXX";

  if (FALSE == links_on_flag) return;
  links_geom_needs_updating = TRUE;

  gv_begin();
  sprintf(name+5, "%d", s->id);
  gv_delete_handle(name);
  gv_end();
}



void
write_links_geom(const Constellation *pconstellation)
{
  Satellite s;
  Satellite_list sl = pconstellation->satellites;
  char buf[256];

  if ((links_on_flag) && (links_geom_needs_updating == TRUE))
    {
      /* write out new list of all satellites */
      gv_send("(geometry Links {LIST\n");
      while (NULL != sl)
	{
	  s = sl->s;
	  if (s->can_display_links) {
	    sprintf(buf, "{ INST transform:trans_%d geom:link_%d}\n", s->id, s->id);
	    gv_send(buf);
	  }
	  sl = sl->next;
	}
      gv_send("})\n");
      
      links_geom_exists = TRUE;
      links_geom_needs_updating = FALSE;
    }
}

