/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 by The Geometry Center.
 *  This file is part of SaVi.  SaVi is free software;
 *  you can redistribute it and/or modify it only under
 *  the terms given in the file COPYRIGHT which you should
 *  have received along with this file.  SaVi may be
 *  obtained by emailing savi@geom.umn.edu.
 *
 *****************************************************
 *  
 * tcl_utils.c
 *
 */
#include <tcl.h>
#include <tk.h>

#include "constants.h"
#include "sats.h"
#include "utils.h"
#include "globals.h"
#include "tcl_utils.h"
int Plot_Init(Tcl_Interp *interp);

static Tk_Window mainWindow;	/* The main window for the application. */
Tcl_Interp *interp;	/* Interpreter for this application. */

/* we want to bypass Tcl_Eval for speed in some case, so store some cmd info */
static Tcl_CmdInfo update_info;

/*
 * tk_init
 *
 * Initialization of tcl/tk interface
 */
int
tk_init()
{
  static char title[] = "Satellites";
  static char class[] = ":0";

  interp = Tcl_CreateInterp();
#if ( (TK_MAJOR_VERSION > 4) || ( (TK_MAJOR_VERSION == 4) && (TK_MINOR_VERSION > 0)) )
  if ( !(mainWindow = Tk_CreateWindow(interp, NULL, title, class)))
    {
      error_format("tk_init: Can't create main wainWindow: %s",interp->result);
      return NOT_OK;
    }
#else
  if ( !(mainWindow = Tk_CreateMainWindow(interp, NULL, title, class)))
    {
      error_format("tk_init: Can't create main wainWindow: %s",interp->result);
      return NOT_OK;
    }
#endif

  if (Tcl_Init(interp) == TCL_ERROR) 
    {
      error_format("tk_init: Error in Tcl_Init: %s", interp->result);
      return NOT_OK;
    }
  if (Tk_Init(interp) == TCL_ERROR) {
    {
      error_format("tk_init: Error in Tk_Init: %s", interp->result);
      return NOT_OK;
    }
  }
  if (Plot_Init(interp) == TCL_ERROR) {
    {
      error_format("tk_init: Error in Plot_Init: %s", interp->result);
      return NOT_OK;
    }
  }
  if (tcl_app_init() == NOT_OK)
    {
      error_format("tk_init: Error in app_init: %s", interp->result);
      return NOT_OK;
    }
  tk_update();

  return OK;
}

/*
 * tcl_app_init
 *
 * Application specific initialization commands for tcl/tk
 *
 */
int
tcl_app_init()
{
  CentralBody *pcb = get_constellation()->pcb;
  static char command0[] = "satellites";
  static char command1[] = "savi";
  static char update[] = "update";
  static char var0[] = "delta_t";
  static char var1[] = "coverage_angle";
  static char var2[] = "IM_W";
  static char var3[] = "IM_H";
  static char var4[] = "params(Radius)";
  static char var5[] = "params(Mu)";
  static char var6[] = "params(Omega)";
  static char var7[] = "params(J2)";
  static char var8[] = "tracks_interval";

  /* add directory of tcl code */
  prepend_to_auto_path(TCL_DIR);

  /* add special commands */
  Tcl_CreateCommand(interp, command0, satCmd, (ClientData) NULL,
		     (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, command1, satCmd, (ClientData) NULL,
		     (Tcl_CmdDeleteProc *) NULL);

  /* get info on some commands for optimization */
  Tcl_GetCommandInfo(interp, update, &update_info);

  /* link some variables */
  Tcl_LinkVar(interp, var0, (char *) &delta_t, TCL_LINK_DOUBLE);
  Tcl_LinkVar(interp, var1, (char *) &coverage_angle, TCL_LINK_DOUBLE);
  Tcl_LinkVar(interp, var2, (char *) &Image_Width, TCL_LINK_INT);
  Tcl_LinkVar(interp, var3, (char *) &Image_Height, TCL_LINK_INT);
  Tcl_LinkVar(interp, var4, (char *) &(pcb->radius), TCL_LINK_DOUBLE);
  Tcl_LinkVar(interp, var5, (char *) &(pcb->mu), TCL_LINK_DOUBLE);
  Tcl_LinkVar(interp, var6, (char *) &(pcb->rotation_rate), TCL_LINK_DOUBLE);
  Tcl_LinkVar(interp, var7, (char *) &(pcb->J2), TCL_LINK_DOUBLE);
  Tcl_LinkVar(interp, var8, (char *) &tracks_interval, TCL_LINK_DOUBLE);

  /* execute initialization */
  {
    static char buf0_0[] = "set geomview_module 0";
    static char buf0_1[] = "set geomview_module 1";
    static char buf1[] = "app_init";

    if (geomview_module) tcl_script(buf0_1);
    else tcl_script(buf0_0);
    tcl_script(buf1);
  }

  return OK;
}

/*
 * autopath is the directory list variable which tcl uses to 
 * look for unknown commands
 */
int
prepend_to_auto_path(const char path[])
{
  char buf[256];

  sprintf(buf, "set auto_path [linsert $auto_path 0 %s]", path);
  return tcl_script(buf);
}


/*
 * tcl_script
 *
 * execute script in tcl
 */
int
tcl_script(char script[])
{
  if (Tcl_GlobalEval(interp,script) == TCL_ERROR)
    {
      error_format("tcl_error: %s", interp->result);
      return NOT_OK;
    }
  return OK;
}


/*
 * tk_update
 *
 * Routine to process tk events
 *
 */
int
tk_update()
{
  static char update[] = "update";
  char *argv[1];
  int argc=1;

  argv[0] = update;
  (*update_info.proc)(update_info.clientData, interp, argc, argv);

#if ( (TK_MAJOR_VERSION > 4) || ( (TK_MAJOR_VERSION == 4) && (TK_MINOR_VERSION > 0)) )
  if (Tk_GetNumMainWindows() > 0) return OK;
#else
  if (tk_NumMainWindows > 0) return OK;
#endif
  return NOT_OK;
}


/*
 * returns ClientData for a tcl object
 */
ClientData
get_ClientData(char *obj)
{
  Tcl_CmdInfo info;

  if (Tcl_GetCommandInfo(interp, obj, &info))
    return info.clientData;

  fprintf(stderr,"get_ClientData - bad object name %s\n",obj);
  return NULL;
}

